﻿local addonName, addon = ...
_G[addonName] = addon


-------------------------------------------------------------------------------
-- WoW API
-------------------------------------------------------------------------------
local ChatFrame_AddMessageEventFilter = _G.ChatFrame_AddMessageEventFilter
local ChatFrame_RemoveMessageEventFilter = _G.ChatFrame_RemoveMessageEventFilter
local ClearOverrideBindings = _G.ClearOverrideBindings
local CreateFrame = _G.CreateFrame
local GetAddOnMetadata = _G.GetAddOnMetadata
local GetBindingKey = _G.GetBindingKey
local GetRaidTargetIndex = _G.GetRaidTargetIndex
local GetSpellInfo = _G.GetSpellInfo
local InCombatLockdown = _G.InCombatLockdown
local IsFlying = _G.IsFlying
local IsInGroup = _G.IsInGroup
local IsInRaid = _G.IsInRaid
local IsPlayerSpell = _G.IsPlayerSpell
local SetOverrideBinding = _G.SetOverrideBinding
local SetRaidTarget = _G.SetRaidTarget
local UnitBattlePetLevel = _G.UnitBattlePetLevel
local UnitExists = _G.UnitExists
local UnitGUID = _G.UnitGUID
local UnitIsBattlePet = _G.UnitIsBattlePet
local UnitIsDead = _G.UnitIsDead
local UnitName = _G.UnitName


-------------------------------------------------------------------------------
-- Lua API
-------------------------------------------------------------------------------
local format = _G.format
local pairs = _G.pairs
local select = _G.select
local strlower = _G.strlower
local strtrim = _G.strtrim
local tinsert = _G.tinsert
local tonumber = _G.tonumber


-------------------------------------------------------------------------------
-- Local data
-------------------------------------------------------------------------------
local addonNameUpper = addonName:upper()
addon.title = GetAddOnMetadata(addonName, "Title")

local eventHandler

local clickButton
local clickButtonName = addonName.."Button"
local clickButtonCommand = "CLICK "..clickButtonName..":LeftButton"

local outerButton
local outerButtonName = clickButtonName.."Outer"
local outerButtonCheckTemplate = "/click "..outerButtonName.." check%d\n"
local outerButtonElectTemplate = "/click "..outerButtonName.." elect"

local innerButton
local innerButtonName = clickButtonName.."Inner"
local innerButtonCheckTemplate = "/tar %s\n/click [exists] "..innerButtonName.." check\n"
local innerButtonElectTemplate = "/tar %s\n/click [exists] "..innerButtonName.." elect"

addon.PET_TRACKER = "PetTracker"
local PET_TRACKER = addon.PET_TRACKER
local PET_TRACKER_FILTER = "PetTrackerMapFilter"

local MACRO_CHAR_LIMIT = 1024
local checkMacros = {}
local targets = {}
local minRange
local nearestTarget

local killMacro

local UNCOMMON = 3
local quality
local battlepet
local forfeit
local clicked = false
local raidTargetIndex = {}

local target
--local icon = "|TInterface\\Icons\\Tracking_WildPet:0|t"

-- Database
local db


-------------------------------------------------------------------------------
-- Global data
-------------------------------------------------------------------------------
_G["BINDING_HEADER_"..addonNameUpper] = addon.title
_G["BINDING_NAME_"..clickButtonCommand] = CATCH_PET


-------------------------------------------------------------------------------
-- Global functions
-------------------------------------------------------------------------------
local IsInBattle = C_PetBattles.IsInBattle
local IsWildBattle = C_PetBattles.IsWildBattle
local GetBreedQuality = C_PetBattles.GetBreedQuality


-------------------------------------------------------------------------------
-- Libs
-------------------------------------------------------------------------------
local rc = LibStub("LibRangeCheck-2.0")


-------------------------------------------------------------------------------
-- Pet Battle functions
-------------------------------------------------------------------------------
local function Quality(index)
	return GetBreedQuality(LE_BATTLE_PET_ENEMY, index)
end

local function MaxQuality()
	return max(Quality(1), Quality(2), Quality(3))
end


-------------------------------------------------------------------------------
-- Target functions
-------------------------------------------------------------------------------

-- Removes all RAID markers of any target.
local function ClearTargetMarker()
	local changed
	for i=1,8 do
		if raidTargetIndex[i] then
			raidTargetIndex[i] = nil
			SetRaidTarget("player", i)
			changed = true
		end
	end
	if changed then
		SetRaidTarget("player", 0)
	end
end

-- Marks the current target with the specified marker.
local function SetTargetMarker(index)
	local i = GetRaidTargetIndex("target")
	if i ~= index then
		SetRaidTarget("target", index)
		raidTargetIndex[i or 0] = nil
		raidTargetIndex[index] = true
	end
end

-- Marks the current target depending of its quality.
local function SetQualityMarker()
	ClearTargetMarker()
	if target ~= battlepet then
		SetTargetMarker(7)
	elseif quality < UNCOMMON then
		SetTargetMarker(8)
	elseif quality > UNCOMMON then
		SetTargetMarker(1)
	else
		SetTargetMarker(4)
	end
end

-- Returns the guid of the current target, if it is a living battle pet.
local function TargetGUID()
	if UnitIsBattlePet("target") and not UnitIsDead("target") then
		local petName = strlower(UnitName("target") or "")
		local petLevel = UnitBattlePetLevel("target") or 1
		for name, level in pairs(targets) do
			if petName:find(name) == 1 then
				if petLevel >= level then
					return UnitGUID("target")
				end
				break
			end
		end
	end
end


-------------------------------------------------------------------------------
-- Key functions
-------------------------------------------------------------------------------

-- Filters wrong error messages from the chat, if the user spams the key.
local function ChatFilter(self, event, ...)
	local message = ...
	if message == ERR_PETBATTLE_IN_BATTLE then
		return not IsInBattle()
	end
end

-- Removes the temporary override binding.
local function ClearInteractTargetKey()
	local key = GetBindingKey(clickButtonCommand)
	if key then
		ChatFrame_RemoveMessageEventFilter("CHAT_MSG_SYSTEM", ChatFilter)
		ClearOverrideBindings(clickButton)
	end
end

-- Sets the temporary override binding.
local function SetInteractTargetKey()
	local key1, key2 = GetBindingKey(clickButtonCommand)
	if key1 then
		ChatFrame_AddMessageEventFilter("CHAT_MSG_SYSTEM", ChatFilter)
		SetOverrideBinding(clickButton, true, key1, "INTERACTTARGET")
		if key2 then
			SetOverrideBinding(clickButton, true, key2, "INTERACTTARGET")
		end
	end
end


-------------------------------------------------------------------------------
-- Macro functions
-------------------------------------------------------------------------------

-- Takes the filtered species of the current zone from PetTracker.
local function GetZoneSpecies()
	local text = ""
	local tracker = _G[PET_TRACKER]
	if tracker then
		local journal = tracker.Journal
		local species = journal:GetSpeciesIn(tracker.zone)
		for specie, floors in pairs(species) do
			if floors[tracker.level] and tracker:Filter(tracker.Specie:Get(specie), tracker.Sets.MapFilter) then
				local name, _, _, _, _, _, isWild = journal:GetInfo(specie)
				if isWild then
					text = text..","..journal:GetInfo(specie)
				end
			end
		end
	end
	return text:sub(2)
end

-- Updates the INNER macros.
local function UpdateInnerMacros(text)
	if text and text ~= "" then
		wipe(checkMacros)
		wipe(targets)

		local checkMacro = ""

		for name, level in text:gmatch("([^,:]+):*(%d*)") do
			name = strlower(strtrim(name))
			if name ~= "" and not targets[name] then
				targets[name] = max(1, min(25, tonumber(level) or 1))
				local macro = format(innerButtonCheckTemplate, name)
				if checkMacro:len() + macro:len() > MACRO_CHAR_LIMIT then
					tinsert(checkMacros, checkMacro)
					checkMacro = ""
				end
				checkMacro = checkMacro..macro
			end
		end
		tinsert(checkMacros, checkMacro)
	end
end

-- Updates the FIND macro.
function addon:UpdateFindMacro()
	if db.options.usePetTracker then
		local filter = _G[PET_TRACKER_FILTER]
		if filter and not filter.OnTextChanged then
			filter.OnTextChanged = filter:GetScript("OnTextChanged")
			filter:SetScript("OnTextChanged", function(self, ...)
				self:OnTextChanged(...)
				UpdateInnerMacros(GetZoneSpecies())
			end)
			eventHandler:RegisterEvent("ZONE_CHANGED_NEW_AREA")
		end
		UpdateInnerMacros(GetZoneSpecies())
	else
		local filter = _G[PET_TRACKER_FILTER]
		if filter and filter.OnTextChanged then
			filter:SetScript("OnTextChanged", filter.OnTextChanged)
			filter.OnTextChanged = nil
			eventHandler:UnregisterEvent("ZONE_CHANGED_NEW_AREA")
		end
		UpdateInnerMacros(db.options.findMacroText)
	end
end

-- Updates the KILL macro.
function addon:UpdateKillMacro()
	killMacro = ""
	local text = db.options.killMacroText
	if text and text ~= "" then
		for name in text:gmatch("[^,]+") do
			name = strlower(strtrim(name))
			if name ~= "" then
				if name == SLASH_STARTATTACK1:sub(2) then
					killMacro = killMacro..SLASH_STARTATTACK1.."\n"
				elseif name == SLASH_STARTATTACK2:sub(2) then
					killMacro = killMacro..SLASH_STARTATTACK2.."\n"
				else
					local spellID = select(7, GetSpellInfo(name))
					if spellID and IsPlayerSpell(spellID) then
						killMacro = killMacro.."/cast [harm] "..name.."\n"
					end
				end
			end
		end
	end
end


-------------------------------------------------------------------------------
-- Initialization functions
-------------------------------------------------------------------------------

-- Initializes the addon.
function addon:Initialize()
	db = self:InitDatabase()

	self:CreateClickButton()
	self:CreateOuterButton()
	self:CreateInnerButton()

	self:CreateSlashHandler()
	self:CreateInterfacePanels()

	self:UpdateFindMacro()
	self:UpdateKillMacro()
end


-------------------------------------------------------------------------------
-- Creates and initializes the CLICK button
-------------------------------------------------------------------------------
function addon:CreateClickButton()
	local frame = CreateFrame("Button", clickButtonName, UIParent, "SecureHandlerBaseTemplate SecureActionButtonTemplate ActionButtonTemplate")
	if db.options.hideButton then
		frame:Hide()
	end
	frame.icon:SetTexture([[Interface/Icons/Tracking_WildPet]])
	frame:SetAttribute("type1", "macro")
	frame:SetClampedToScreen(true)
	frame:SetMovable(true)
	frame:SetPoint(db.anchor.point, nil, db.anchor.relativePoint, db.anchor.xOffset, db.anchor.yOffset)
	frame:SetScript("OnMouseDown", function(self, button)
		if button == "RightButton" then
			self:StartMoving()
		end
	end)
	frame:SetScript("OnMouseUp", function(self, button)
		if button == "RightButton" then
			self:StopMovingOrSizing()
			self:SetUserPlaced(false)
			local _
			db.anchor.point, _, db.anchor.relativePoint, db.anchor.xOffset, db.anchor.yOffset = self:GetPoint()
		end
	end)

	function frame:PreClick(button)
		if IsInBattle() then
			if not forfeit then
				C_PetBattles.ForfeitGame()
				forfeit = true
			end
		elseif not InCombatLockdown() then
			local macrotext = ""
			if not target then
				for i=1,#checkMacros do
					macrotext = macrotext..format(outerButtonCheckTemplate, i)
				end
				minRange = db.options.maxDistance
				nearestTarget = nil
			end
			macrotext = macrotext..outerButtonElectTemplate
			self:SetAttribute("macrotext", macrotext)
			self:SetAttribute("newbutton", nil)
			clicked = true
		end
	end

	function frame:PostClick()
		clicked = false
	end

	frame:WrapScript(frame, "OnClick", [[
		self:SetAttribute("newbutton", false)
		self:CallMethod("PreClick", button)
		return self:GetAttribute("newbutton"), true
	]], [[
		self:CallMethod("PostClick")
	]])

	self.clickButton = frame
	clickButton = frame
end


-------------------------------------------------------------------------------
-- Creates and initializes the OUTER button
-------------------------------------------------------------------------------
function addon:CreateOuterButton()
	local frame = CreateFrame("Button", outerButtonName, nil, "SecureHandlerBaseTemplate SecureActionButtonTemplate")
	frame:Hide()
	frame:SetAttribute("type", "macro")

	function frame:PreClick(button)
		if button then
			if button == "elect" then
				if not target then
					if nearestTarget then
						self:SetAttribute("macrotext", format(innerButtonElectTemplate, nearestTarget))
						self:SetAttribute("newbutton", nil)
					end
				elseif killMacro and killMacro ~= "" then
					if not IsFlying() and GetRaidTargetIndex("target") == 8 then
						self:SetAttribute("macrotext", killMacro)
						self:SetAttribute("newbutton", nil)
					end
				end
			elseif button:sub(1,5) == "check" then
				local i = tonumber(button:sub(6))
				if i then
					self:SetAttribute("macrotext", checkMacros[i])
					self:SetAttribute("newbutton", nil)
				end
			end
		end
	end

	frame:WrapScript(frame, "OnClick", [[
		--print(self:GetName(), "OnClick", button)
		self:SetAttribute("newbutton", false)
		self:CallMethod("PreClick", button)
		return self:GetAttribute("newbutton")
	]])

	outerButton = frame
end


-------------------------------------------------------------------------------
-- Creates and initializes the INNER button
-------------------------------------------------------------------------------
function addon:CreateInnerButton()
	local frame = CreateFrame("Button", innerButtonName, nil, "SecureHandlerBaseTemplate SecureActionButtonTemplate")
	frame:Hide()
	frame:SetAttribute("type", "macro")

	function frame:PreClick(button)
		if button then
			if button == "check" then
				local target = TargetGUID()
				if target then
					local min = rc:GetRange("target")
					if min < minRange or min == minRange and not nearestTarget then
						minRange = min
						nearestTarget = UnitName("target")
					end
				end
				if UnitExists("target") then
					self:SetAttribute("macrotext", "/cleartarget")
					self:SetAttribute("newbutton", nil)
				end
			elseif button == "elect" then
				target = TargetGUID()
				if target then
					if target == battlepet and (quality < UNCOMMON or GetRaidTargetIndex("target") == 4) then
						SetTargetMarker(8)
					else
						SetQualityMarker()
						SetInteractTargetKey()
					end
				else
					self:SetAttribute("macrotext", "/cleartarget")
					self:SetAttribute("newbutton", nil)
				end
			end
		end
	end

	frame:WrapScript(frame, "OnClick", [[
		self:SetAttribute("newbutton", false)
		self:CallMethod("PreClick", button)
		return self:GetAttribute("newbutton")
	]])

	innerButton = frame
end


-------------------------------------------------------------------------------
-- Creates the slash handler
-------------------------------------------------------------------------------
function addon:CreateSlashHandler()
	_G["SLASH_"..addonNameUpper..1] = "/pf"
	_G["SLASH_"..addonNameUpper..2] = "/petfarmer"

	SlashCmdList[addonNameUpper] = function(msg)
		if InCombatLockdown() then
			return
		end
		if msg == "help" then
			print(format("|cff33ff99%s %s", addonName, HELP_LABEL))
			print("  /pf |cff33ff99- Opens the options panel|r")
			print("  /pf help |cff33ff99- Shows this help|r")
			print("  /pf find <name[:#]>,... |cff33ff99- Sets a comma separated list of strings (optional with min level) to search battle pets|r")
			print("       |cffffff78Example: /pf find toll,elox:24,robo|r")
			print("  /pf kill <spell>,... |cff33ff99- Sets a comma separated list of spells to kill battle pets|r")
			print("       |cffffff78Example: /pf kill startattack,throw,ice lance|r")
			print("  You can also type /petfarmer instead of /pf|r")
			print("\n")
		elseif msg:sub(1,5) == "find " then
			if msg:sub(6) ~= db.options.findMacroText then
				db.options.findMacroText = msg:sub(6)
				self:UpdateFindMacro()
			end
		elseif msg:sub(1,5) == "kill " then
			if msg:sub(6) ~= db.options.killMacroText then
				db.options.killMacroText = msg:sub(6)
				self:UpdateKillMacro()
			end
		elseif msg == "" then
			InterfaceOptionsFrame_OpenToCategory(addonName)
			InterfaceOptionsFrame_OpenToCategory(addonName)
		end
	end
end


-------------------------------------------------------------------------------
-- Creates and initializes the event handler frame
-------------------------------------------------------------------------------
do
	local frame = CreateFrame("Frame")

	frame:RegisterEvent("PLAYER_LOGIN")
	frame:RegisterEvent("PET_BATTLE_OPENING_START")
	frame:RegisterEvent("PLAYER_TARGET_CHANGED")
	frame:RegisterEvent("PLAYER_REGEN_ENABLED")

	frame:SetScript("OnEvent", function(self, event, ...)
		if event == "PLAYER_LOGIN" then
			self:UnregisterEvent(event)
			addon:Initialize()
			return
		end
		if IsInRaid() or IsInGroup() or InCombatLockdown() then
			return
		end
		if event == "ZONE_CHANGED_NEW_AREA" then
			--print(event, ...)
			UpdateInnerMacros(GetZoneSpecies())
			return
		end
		if clicked or not target then
			return
		end
		if event == "PET_BATTLE_OPENING_START" then
			--print(event, ...)
			ClearInteractTargetKey()
			forfeit = nil
			battlepet = IsWildBattle() and target
			if battlepet then
				quality = MaxQuality()
				SetQualityMarker()
			end
		elseif event == "PLAYER_TARGET_CHANGED" then
			--print(event, (UnitName("target")))
			ClearInteractTargetKey()
			if not IsInBattle() then
				ClearTargetMarker()
			end
			target = nil
		elseif event == "PLAYER_REGEN_ENABLED" then
			--print(event, ...)
			ClearInteractTargetKey()
			ClearTargetMarker()
			target = nil
		end
	end)

	eventHandler = frame
end


function addon:Info()
	print("\n")
	print("target:", target)
	print("battlepet:", battlepet)
	print("clicked:", clicked)
	print("binding:", GetBindingAction(GetBindingKey(clickButtonCommand) or "", true))
	print("IsInBattle:", IsInBattle())
	print("#checkMacros:", #checkMacros)
	print("clickButton:", clickButton:GetAttribute("macrotext"))
	print("outerButton:", outerButton:GetAttribute("macrotext"))
	print("\n")
end
